# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sns_sqs_message.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15sns_sqs_message.proto\x12\x05proto\"%\n\x07Service\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04uuid\x18\x02 \x01(\t\"s\n\x08Metadata\x12\x14\n\x0cmessage_uuid\x18\x01 \x01(\t\x12\x18\n\x10protocol_version\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\x01\x12\r\n\x05topic\x18\x04 \x01(\t\x12\x15\n\rdata_encoding\x18\x05 \x01(\t\"a\n\rSNSSQSMessage\x12\x1f\n\x07service\x18\x01 \x01(\x0b\x32\x0e.proto.Service\x12!\n\x08metadata\x18\x02 \x01(\x0b\x32\x0f.proto.Metadata\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sns_sqs_message_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SERVICE._serialized_start=32
  _SERVICE._serialized_end=69
  _METADATA._serialized_start=71
  _METADATA._serialized_end=186
  _SNSSQSMESSAGE._serialized_start=188
  _SNSSQSMESSAGE._serialized_end=285
# @@protoc_insertion_point(module_scope)
